"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DesignType_1 = require("./DesignType");
const BYTE_LENGTH_NORMAL = 620;
const BYTE_LENGTH_PRO = 540;
class QRData {
    constructor(type, index) {
        // Make sure we have valid parameters
        if (type === DesignType_1.DesignType.Normal) {
            if (index !== undefined) {
                throw new Error('Index must be set to "undefined" for normal designs');
            }
        }
        else if (type === DesignType_1.DesignType.Pro) {
            if (index === undefined || index < 0 || index > 3) {
                throw new Error('Must specify a QR code index between 0-3 for pro designs');
            }
        }
        this.type = type;
        this.index = index;
        this.data = this.getNewBuffer();
    }
    static fromBytes(data, index) {
        let result;
        if (data.length === BYTE_LENGTH_NORMAL) {
            result = new QRData(DesignType_1.DesignType.Normal);
        }
        else if (data.length === BYTE_LENGTH_PRO) {
            result = new QRData(DesignType_1.DesignType.Pro, index);
        }
        else {
            throw new Error('Invalid data length');
        }
        return result.loadBytes(data);
    }
    // Properties
    /**
     * The design's title (UTF-16 encoded, up to 42 bytes).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get title() {
        return this.hasMetadata() ? this.getString(0, 42) : undefined;
    }
    set title(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setString(0, 42, value);
    }
    /**
     * The player ID (2 bytes).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get playerId() {
        return this.hasMetadata() ? this.getNumber(42, 2) : undefined;
    }
    set playerId(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(42, 2, value);
    }
    /**
     * The creator's name (UTF-16 encoded, up to 18 bytes).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get creator() {
        return this.hasMetadata() ? this.getString(44, 18) : undefined;
    }
    set creator(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setString(44, 18, value);
    }
    /**
     * The creator's sex (1 byte).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get sex() {
        return this.hasMetadata() ? this.getNumber(62, 1) : undefined;
    }
    set sex(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(62, 1, value);
    }
    /**
     * The village ID (2 bytes).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get villageId() {
        return this.hasMetadata() ? this.getNumber(64, 2) : undefined;
    }
    set villageId(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(64, 2, value);
    }
    /**
     * The village name (UTF-16 encoded, up to 18 bytes).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get village() {
        return this.hasMetadata() ? this.getString(66, 18) : undefined;
    }
    set village(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setString(66, 18, value);
    }
    /**
     * The language ID (1 byte).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get language() {
        return this.hasMetadata() ? this.getNumber(84, 1) : undefined;
    }
    set language(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(84, 1, value);
    }
    /**
     * The country ID (1 byte).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get country() {
        return this.hasMetadata() ? this.getNumber(86, 1) : undefined;
    }
    set country(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(86, 1, value);
    }
    /**
     * The region ID (1 byte).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get region() {
        return this.hasMetadata() ? this.getNumber(87, 1) : undefined;
    }
    set region(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(87, 1, value);
    }
    /**
     * The palette colors (15 bytes).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get paletteColors() {
        return this.hasMetadata() ? this.getByteSubset(88, 15) : undefined;
    }
    set paletteColors(value) {
        this.throwIfInvalidMetadataSetter(value);
        if (value.length !== 15) {
            throw new Error('Invalid data length');
        }
        this.setByteSubset(88, value);
    }
    /**
     * The color (this field's purpose is currently unknown) (1 byte).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get color() {
        return this.hasMetadata() ? this.getNumber(103, 1) : undefined;
    }
    set color(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(103, 1, value);
    }
    /**
     * The looks (this field's purpose is currently unknown) (1 byte).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get looks() {
        return this.hasMetadata() ? this.getNumber(104, 1) : undefined;
    }
    set looks(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(104, 1, value);
    }
    /**
     * The usage (1 byte).
     * Only supported for normal design QR codes and the first QR code for pro designs.
     */
    get usage() {
        return this.hasMetadata() ? this.getNumber(105, 1) : undefined;
    }
    set usage(value) {
        this.throwIfInvalidMetadataSetter(value);
        this.setNumber(105, 1, value);
    }
    /**
     * Each pixel's color data.
     * For pro designs, this only contains a segment of the full design's pixel data.
     * Total size varies depending on QR code type and index.
     */
    get colorData() {
        return this.getByteSubset(...this.colorDataByteRange());
    }
    set colorData(value) {
        const range = this.colorDataByteRange();
        if (value.length !== range[1]) {
            throw new Error('Invalid data length');
        }
        this.setByteSubset(range[0], value);
    }
    colorDataByteRange() {
        if (this.type === DesignType_1.DesignType.Normal) {
            return [108, 512];
        }
        switch (this.index) {
            case 0:
                return [108, 432];
            case 1:
            case 2:
                return [0, 540];
            case 3:
                return [0, 536];
        }
        throw new Error('Invalid parameters');
    }
    // Byte data
    loadBytes(data) {
        if (data.length !== this.data.length) {
            throw new Error('Invalid data length');
        }
        this.data.set(data);
        return this;
    }
    getBytes() {
        return this.data;
    }
    // Other helpers
    hasMetadata() {
        return (this.type === DesignType_1.DesignType.Normal ||
            (this.type === DesignType_1.DesignType.Pro && this.index === 0));
    }
    throwIfInvalidMetadataSetter(value) {
        if (!this.hasMetadata()) {
            throw new Error('This QR data type has no metadata');
        }
        if (value === undefined || value === null) {
            throw new Error('Value is required');
        }
    }
    getNewBuffer() {
        if (this.type === DesignType_1.DesignType.Normal) {
            return new Uint8Array(BYTE_LENGTH_NORMAL);
        }
        else if (this.type === DesignType_1.DesignType.Pro) {
            return new Uint8Array(BYTE_LENGTH_PRO);
        }
        throw new Error('Invalid design type');
    }
    getByteSubset(offset, length) {
        return this.data.subarray(offset, offset + length);
    }
    getString(offset, length) {
        const bytes = this.getByteSubset(offset, length);
        const result = new TextDecoder('utf-16le', { ignoreBOM: true }).decode(bytes);
        return result.includes('\u0000') ? result.split('\u0000', 1)[0] : result;
    }
    getNumber(offset, length) {
        return this.getByteSubset(offset, length)
            .reduce((result, b) => (result << 8) + b, 0);
    }
    setByteSubset(offset, data) {
        this.data.set(data, offset);
    }
    setString(offset, length, value) {
        // Convert the string to UTF-16 LE
        // NOTE: This may not truncate multi-code unit characters properly. Code points
        // that require more than 2 bytes may be malformed.
        const bytes = new Uint8Array(length);
        for (let i = 0; i < value.length && i * 2 < bytes.length; i++) {
            const charCode = value.charCodeAt(i);
            bytes[i * 2] = charCode;
            bytes[i * 2 + 1] = charCode >> 8;
        }
        this.setByteSubset(offset, bytes);
    }
    setNumber(offset, length, value) {
        for (let i = 0; i < length; i++) {
            this.data[offset + i] = value >> (8 * (length - i - 1));
        }
    }
}
exports.QRData = QRData;
